/* Pound Includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mex.h"
#include "matrix.h"
#include "cuda.h"
#include "cublas.h"

/* Pound Defines  */
#define NumberThreadsPerBlock 128


/* ========== */
/* GPU Kernel */
/* ========== */

/* ============ */
/* Main Program */
/* ============ */

// Matlab Main
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
	// Declarations
    mwSize InputRowsX;
    mwSize InputColsX;
    mwSize InputRowsY;
    mwSize InputColsY;
    mwSize InputRowsW;
    mwSize InputColsW;
    mwSize InputRowsP;
    mwSize InputColsP;
    mwSize NumParams;
    mwSize NumData;
    mwIndex irow;
    mwIndex icol;
    mwIndex count = 0;
    double norm_res;
    double *xdata;
    double *ydata;
    double *weights;
    double *p;
    double *xdata_gpu;
    double *ydata_gpu;
    double *weights_gpu;
    double *A_gpu;
    double *b_gpu;
    double *Ab_gpu;
    double *D_gpu;
    double *Dinv_gpu;
    double *p_gpu;
    double *residual_gpu;
    cublasStatus status;
    
	// Function Prototypes

	/* ============== */
	/* Input Checking */
	/* ============== */

	// Initialize CUBLAS
    status = cublasInit();
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Initialization Error\n");
    
    // Check Number of Inputs
	// if (nrhs >3)
	// 	mexErrMsgTxt("Matlab_CUDA_DFT only recongizes three input arguments: X, n, dim");
	// if (nlhs !=1)
	// 	mexErrMsgTxt("Matlab_CUDA_DFT only recongizes one output argument: Y");

    // Get Input Sizes
	InputRowsX = mxGetM( prhs[0]);
	InputColsX = mxGetN( prhs[0]);
	InputRowsY = mxGetM( prhs[1]);
	InputColsY = mxGetN( prhs[1]);
	InputRowsW = mxGetM( prhs[2]);
	InputColsW = mxGetN( prhs[2]);
	InputRowsP = mxGetM( prhs[3]);
	InputColsP = mxGetN( prhs[3]);
    NumParams  = InputRowsP;
    NumData    = InputRowsY;
    
    // Test Inputs
    // if( !mxIsDouble(prhs[0]) || mxIsComplex(prhs[0]) || !(mrows==1 && ncols==1) )
        // mexErrMsgTxt("Input must be a noncomplex scalar double.");
  
    // Create Output Matrix
    plhs[0] = mxCreateDoubleMatrix(InputRowsP,InputColsP, mxREAL);

    // Assign Pointers to Each Input
    xdata   = mxGetPr(prhs[0]);
    ydata   = mxGetPr(prhs[1]);
    weights = mxGetPr(prhs[2]);
    p       = mxGetPr(prhs[3]);

    // Assign Pointers to Each Output
    p       = mxGetPr(plhs[0]);

    // Create Identity Matrix
	double *eye = (double *) mxMalloc( sizeof(double)*NumParams*NumParams );
    for (irow=0; irow<NumParams; ++irow)
        for (icol=0; icol<NumParams; ++icol)
            if (irow==icol)
                eye[irow*NumParams+icol] = 1.0;
            else
                eye[irow*NumParams+icol] = 0.0;
    
    // Create an input and output data array on the GPU
    status = cublasAlloc (InputRowsX*InputColsX, sizeof(double), (void **) &xdata_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  xdata_gpu\n");
    
    status = cublasAlloc (InputRowsW*InputColsW, sizeof(double), (void **) &weights_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  weights_gpu\n");

    status = cublasAlloc (NumParams*NumData,     sizeof(double), (void **) &Ab_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  Ab_gpu\n");

    status = cublasAlloc (InputRowsY*InputColsY,     sizeof(double), (void **) &ydata_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  ydata_gpu\n");

    status = cublasAlloc (NumParams*NumParams,     sizeof(double), (void **) &A_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  A_gpu\n");

    status = cublasAlloc (NumParams,     sizeof(double), (void **) &b_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  b_gpu\n");

    status = cublasAlloc (NumParams*NumParams,     sizeof(double), (void **) &D_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  D_gpu\n");

    status = cublasAlloc (NumParams*NumParams,     sizeof(double), (void **) &Dinv_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  Dinv_gpu\n");

    status = cublasAlloc (InputRowsP*InputColsP,     sizeof(double), (void **) &p_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  p_gpu\n");

    status = cublasAlloc (InputRowsP*InputColsP,     sizeof(double), (void **) &residual_gpu);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Memory Allocation Error:  residual_gpu\n");

    // Copy InputData to GPU Card
    status = cublasSetMatrix (InputRowsX, InputColsX, sizeof(double), xdata, InputRowsX, xdata_gpu, InputRowsX);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Set Matrix Error:  xdata\n");

    status = cublasSetMatrix (InputRowsW, InputColsW, sizeof(double), weights, InputRowsW, weights_gpu, InputRowsW);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Set Matrix Error:  weights\n");

    status = cublasSetMatrix (InputRowsY, InputColsY, sizeof(double), ydata, InputRowsY, ydata_gpu, InputRowsY);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Set Matrix Error:  ydata\n");

    status = cublasSetMatrix (NumParams, NumParams, sizeof(double), eye, NumParams, Dinv_gpu, NumParams);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Set Matrix Error:  eye\n");

    status = cublasSetMatrix (NumParams, NumParams, sizeof(double), eye, NumParams, D_gpu, NumParams);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Set Matrix Error:  D\n");

    status = cublasSetMatrix (InputRowsP, InputColsP, sizeof(double), p, InputRowsP, p_gpu, InputRowsP);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Set Matrix Error:  p\n");

    // Call CUBLAS to do Matrix Multiply
    cublasDgemm( 't', 'n', NumParams, NumData, NumData, 1.0, xdata_gpu, NumData, weights_gpu, NumData, 0.0, Ab_gpu, NumParams);
    cublasDgemm( 'n', 'n', NumParams, NumParams, NumData, 1.0, Ab_gpu, NumParams, xdata_gpu, NumData, 0.0, A_gpu, NumParams);
    cublasDgemv( 'n', NumParams, NumData, 1.0, Ab_gpu, NumParams, ydata_gpu, 1, 0.0, b_gpu, 1);
    cublasDcopy( NumParams*NumParams, A_gpu, NumParams+1, D_gpu, NumParams+1);
    cublasDtrsm ( 'l', 'l', 'n', 'n', NumParams, NumParams, 1.0, D_gpu, NumParams, Dinv_gpu, NumParams);

    // Copy result back to host
 	double *Dinv = (double *) mxMalloc( sizeof(double)*NumParams*NumParams );
    status = cublasGetMatrix (NumParams, NumParams, sizeof(double), Dinv_gpu, NumParams, Dinv, NumParams);
    if (status != CUBLAS_STATUS_SUCCESS)
        mexPrintf("CUBLAS Get Matrix Error:  Dinv_gpu\n");
    mwIndex ii;
    for (ii=0; ii<NumParams*NumParams; ++ii)
        mexPrintf("D = %e\n",Dinv[ii]);

//     cublasDcopy( NumParams*NumParams, b_gpu, 1, residual_gpu, 1);
//     cublasDgemv( 'n', NumParams, NumParams, -1.0, A_gpu, NumParams, p_gpu, 1, 1.0, residual_gpu, 1);
//     norm_res = cublasDnrm2 ( NumParams, residual_gpu, 1);
//     
//     while (norm_res > 1e-5)
//     {
//         count+=1;
//         cublasDgemv( 'n', NumParams, NumParams, 1.0, Dinv_gpu, NumParams, residual_gpu, 1, 1.0, p_gpu, 1);
//         cublasDcopy( NumParams*NumParams, b_gpu, 1, residual_gpu, 1);
//         cublasDgemv( 'n', NumParams, NumParams, -1.0, A_gpu, NumParams, p_gpu, 1, 1.0, residual_gpu, 1);
//         norm_res = cublasDnrm2 ( NumParams, residual_gpu, 1);
//     }
//     mexPrintf("Count = %d\n",count);

    // Copy result back to host
    cudaMemcpy( p, p_gpu, sizeof(double)*InputRowsP*InputColsP, cudaMemcpyDeviceToHost );

    // Clean-up memory on device
    cublasFree(xdata_gpu);
    cublasFree(ydata_gpu);
    cublasFree(weights_gpu);
    cublasFree(A_gpu);
    cublasFree(b_gpu);
    cublasFree(Ab_gpu);
    cublasFree(D_gpu);
    cublasFree(Dinv_gpu);
    cublasFree(p_gpu);
    cublasFree(residual_gpu);
    cublasShutdown();

    // Clean-up memory on host
    mxFree(eye);

    return;
}


/* EOF */
